


# L16ortho2.py
# Orthogonal; see LamptonSixteen.xls and Dot.py
# Multiple random refitting of fit data, saving resulting sky statistic
# L13ortho2.py incorporates the sky fit graphic as #0 within the multiple random runs.
# With this change I am trying to capture the "jumpy" statistics
# 
# Trying fits to Mangled8 using all 16 polynomials:  sky fit err = 6 ppm; w/noise:  31ppm jumpy.
# Trying fits to Mangled8 using top 14 polynomials:  sky fit err = 6ppm;  w/100um noise: 34ppm jumpy
# Trying fits to Mangled8 using top 13 polynomials:  sky fit err = 7ppm; with noise, 39ppm jumpy.
# Trying fits to Mangled8 using top 12 polynomials:  sky fit err = 161ppm. Yikes
# Trying fits to Mangled8 using top 11 polynomials:  sky fit err = 182ppm.  Yikes.






#-------------- Procedure ----------------
# Adopt an optical model of E-22 with ideal or mangled optics: ADC00.OPT, ADC45.OPT, or Mangled8.OPT.
# With 120-fiducials as Xgoal,Ygoal, do a BEAM FOUR AutoRay (reverse trace) on 120 chief rays, getting {U0,V0}s.
# Save that file as nnnFid-xxxOpt-fit.RAY
# Now load the sky grid file with ~1027 {U0,V0} coordinates filling the field.
# Forward trace that, getting ~1000 {Xf,Yf} focal plane coordinates.
# Save that file as  optfilename-sky.RAY.
# Now set up this program to load and analyze  xxx-fit.RAY for fitting, and xxx-sky.RAY for testing.


import csv
import numpy as np
from numpy import genfromtxt
import matplotlib.pyplot as plt
import math
from scipy.optimize import curve_fit


NPARMS = 16  # NO LUT
PROGNAME = 'L13ortho2.py'
XYradius = 410.0
UVradius = 0.0282

NTRIALS = 100
NBINS = 400   # all my plots go -200...+200ppm, to accommodate worst case 22fid & max noise

rmsppm       =  100  # ppm random fiducial location error in x and in y.
fiducials    = '120Fid-Mangled8-fit.RAY.CSV'  #  '22Fid-ADC00-fit.RAY.CSV'  #  
bigskygrid   = 'Mangled8-sky.RAY.CSV'        # '22Fid-ADC00-sky.RAY.CSV'  #  'Mangled8-sky.RAY.CSV'
optics       = 'Mangled8.OPT'

tadpole = 10000.
plotnote = str(rmsppm) + ' ppm rms fiducials'


#-----sixteen polynomials---------------------
names = ['X trans',
         'Y trans',
         'Magnify',
         'UpDownUp',
         'RightLeftRight',
         'Cubic radial',
         'R-L-R-L-R',
         'U-D-U-D-U',
         'Radial fifth',
         'Roll',
         'higher',
         'higher',
         'Cubic roll',
         'higher',
         'higher',
         'fifth roll']
         
ortho = np.array(
[[    1,      0,      0,      0,      0,      0,     -1,      0,      0,      0,      0,      0], 
 [    1,      0,      0,      0,      0,      0,      1,      0,      0,      0,      0,      0], 
 [    0,      4,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0], 
 [   -3,      0,      9,      0,      0,      0,     -3,      0,      3,      0,      0,      0], 
 [   -3,      0,      9,      0,      0,      0,      3,      0,     -3,      0,      0,      0], 
 [    0,    -16,      0,     24,      0,      0,      0,      0,      0,      0,      0,      0], 
 [    5,      0,    -45,      0,     50,      0,     -5,      0,     15,      0,    -10,      0], 
 [    5,      0,    -45,      0,     50,      0,      5,      0,    -15,      0,     10,      0], 
 [    0,     36,      0,   -144,      0,    120,      0,      0,      0,      0,      0,      0], 
 [    0,      0,      0,      0,      0,      0,     -0,     -4,      0,      0,      0,      0], 
 [    3,      0,     -3,      0,      0,      0,     -3,      0,      9,      0,      0,      0], 
 [   -3,      0,      3,      0,      0,      0,     -3,      0,      9,      0,      0,      0], 
 [    0,      0,      0,      0,      0,      0,      0,    -16,      0,     24,      0,      0], 
 [    5,      0,    -15,      0,     10,      0,      5,      0,    -45,      0,     50,      0], 
 [   -5,      0,     15,      0,    -10,      0,      5,      0,    -45,      0,     50,      0], 
 [    0,      0,      0,      0,      0,      0,      0,     36,      0,   -144,      0,    120]]) 
 
diag = np.array([   1,    1,    8,    6,    6,   16,   10,   10,   24,    8,    6,    6,   16,   10,   10,   24])
coef = np.power(diag, -0.5)
 


# Key trig values:  0 means 0;, 1 means const = 2pi/pi; 2 means cos; 3 means sin.
# Only the radial part is encoded here; the tangential is always the opposite.
trig = [2, 3, 1, 3, 2, 1, 2, 3, 1, 0, 2, 3, 0, 3, 2, 0]

NROWS = 16
NTERMS = 12
NGROUP = 6


def getOrthoDot(irow, jrow):  # evaluates dot products of any two ortho rows
    if trig[irow] * trig[jrow] == 6.:  # sin x cos integrates to zero.
        return 0.
    sumRadial = 0.
    for col1 in range(0, NGROUP):
        for col2 in range(0, NGROUP):
            if (col1 + col2) % 2 == 0:  # is Even
                term = ortho[irow, col1] * ortho[jrow, col2]
                power1 = col1 % NGROUP
                power2 = col2 % NGROUP
                denom = 2. + power1 + power2 
                term = term/denom
                sumRadial += term
    sumTangential = 0.
    for col1 in range(NGROUP, NTERMS):
        for col2 in range(NGROUP, NTERMS):
            if (col1 + col2) % 2 == 0:  # is Even
                term = ortho[irow, col1] * ortho[jrow, col2]
                power1 = col1 % NGROUP
                power2 = col2 % NGROUP
                denom = 2. + power1 + power2 
                term = term/denom        
                sumTangential += term
    sum = sumRadial + sumTangential
    sum *= coef[irow] * coef[jrow]
    factor = 1.
    if trig[irow] in [0, 1]:  # full 2pi/pi integral
        factor = 2.
    return sum*factor


def getLocal(x, y, row):  # args then parms
    # Labframe Cartesian input {x,y}, local frame cartesian output {r,p}.
    # Unit circle coordinates.
    # Output r is along the local radius; output p is perpendicular to the local radius.
    rin = np.sqrt(x**2 + y**2)
    if rin==0:
        c=1.
        s=0.
    else:
        c = x/rin   # cosine
        s = y/rin   # sine
    key = trig[row]
    radial     = np.polynomial.polynomial.polyval(rin, ortho[row, :6])
    tangential = np.polynomial.polynomial.polyval(rin, ortho[row, 6:])
    
    #--produce the results {r, p}---

    if key == 0:
        return 0., coef[row]*tangential
    if key == 1:
        return coef[row]*radial, 0.
    if key == 2:
        return coef[row]*c*radial, coef[row]*s*tangential
    if key == 3:
        return coef[row]*s*radial,  coef[row]*c*tangential
    

def getLab(x, y, which):
    # lab frame cartesian input, lab frame cartesian output.
    if which==0:
        return 1., 0.
    if which==1:
        return 0., 1.
    r = np.sqrt(x**2 + y**2)
    if r==0:
        c=1.
        s=0.
    else:
        c = x/r   # cosine
        s = y/r   # sine
    r, p = getLocal(x, y, which)
    dx = r*c - p*s
    dy = r*s + p*c
    return dx, dy



#---ray generators-------------

def getRaysCSV(rayname):
    # assumes U,V,X,Y are in columns 0,1,3,4
    # all coordinates are derived from ray tracing, hence "rt" suffix
    # global Urt, Vrt, Xrt, Yrt   # creates these in global context.
    try:
        data = genfromtxt(rayname, skip_header=2, delimiter=',')
    except IOError:
        print "Cannot open this file. Quitting."
        quit()
    # print 'Data array shape = ', data.shape
    Urt = data[:,0]/UVradius
    Vrt = data[:,1]/UVradius
    Xrt = data[:,3]/XYradius
    Yrt = data[:,4]/XYradius
    return Urt, Vrt, Xrt, Yrt

def buildSquares(radius, numside):
    # artificial square sky grid within a circular zone
    nsq = numside*numside   
    start = -radius
    space = 2*radius/(numside-1)
    pairs = list()
    count = 0
    for i in range(0, nsq):
        row = i / numside
        col = i % numside
        x = start + col*space
        y = start + row*space
        r = np.sqrt(x*x+y*y)
        if r <= radius:
            pairs.append([x, y])
            count += 1
    print "buildGrid() has created npairs = ", count
    return np.array(pairs)

def buildCircles(radius, ncircles):
    # artificial circular sky grid within a circle
    pairs = list()
    pairs.append([0.,0.])
    count = 1
    for icirc in range(1, ncircles+1):
        daz = 60./icirc
        off = 0.0 if icirc%2==0 else daz/2.0
        r = icirc * radius/ncircles
        for jaz in range(0, 6*icirc):
            a = np.deg2rad(off + jaz*daz)
            x = r * np.cos(a)
            y = r * np.sin(a)
            pairs.append([x,y])
            count += 1
    print ncircles, count
    return np.array(pairs)
    

#---graphics tadpole support------------


def plotXYtadpoles(hhead, vhead, htail, vtail, magnif, fname):
    print "Starting plotXYtadpoles()"
    dh = htail - hhead   # horiz axis
    dv = vtail - vhead   # vertical axis
    h = hhead + magnif*dh
    v = vhead + magnif*dv
    fig = plt.figure(figsize = (6.0, 6.0))          # inches; 6.5=max LaTeX.
    ax = fig.add_axes([0.15, 0.10, 0.80, 0.80])     # about right.  75x75mm on screen.  
    ax.plot(hhead, vhead, 'ko', markersize=4)       # dots
    ax.plot([hhead,h],[vhead,v],'r', linewidth=1.0)  # line segments
    ax.set_xlim(-1.1, +1.1)
    ax.set_ylim(-1.1, +1.1)
    # plt.axis('equal')  # spoils the specified limits!
    ax.set_xlabel('+X')
    ax.set_ylabel('+Y')
    ax.text(-1.05, -0.95, 'TadpoleMag = '+str(int(magnif)), fontsize=8)
    ax.text(-1.05, -1.00, 'Tails: {x,y} from fiducials', fontsize=8)
    ax.text(-1.05, -1.05, 'Heads: ' + fname, fontsize=8)
    plt.title("{X,Y} deviations from sky {U,V}") 
    fig.savefig('XYtadpoles.eps', format='eps', dpi=1200)
    plt.show()
    
    
def plotFitTadpoles(hhead, vhead, htail, vtail, magnif, parms, rmsx, rmsy, fname, note):
    dh = htail - hhead   # horiz axis
    dv = vtail - vhead   # vertical axis
    hmagtail = hhead + magnif*dh
    vmagtail = vhead + magnif*dv
    fig = plt.figure(figsize = (6.5, 6.5))          # 6.5 inches =max LaTeX.
    ax = fig.add_axes([0.1, 0.1, 0.7, 0.7])         # left, bottom, width, height
    ax.plot(hhead, vhead, 'ko', markersize=2) # dots
    ax.plot([hhead,hmagtail],[vhead,vmagtail],'r')  # line segments
    ax.set_xlim(-1.1, +1.1)
    ax.set_ylim(-1.1, +1.1)
    # plt.axis('equal')  # spoils the specified limits!
    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.text(-1.05, -0.98, 'TadpoleMag = '+str(int(magnif)), fontsize=8)
    ax.text(-1.05, -1.03, 'Heads: {x,y} from fiducials', fontsize=8)
    ax.text(-1.05, -1.08, 'Tails: fit to '+fname, fontsize=8)
    ax.text(-1.05, +1.00, note, fontsize=8)
    ax.text(+0.55, -1.08, PROGNAME, fontsize=8)
    plt.title("{X,Y} fit deviations from actual {X,Y}") 
    # now list the best fit parms.....
    for i in range(0, NPARMS):
        sparm = ' {:2d} {:9.6f}'.format(i, parms[i])
        y = 1.15 - 0.07*i
        ax.text(1.11, y, sparm, fontsize=6)
        sx = ' rmsX={:9.6f}'.format(rmsx)
        sy = ' rmsY={:9.6f}'.format(rmsy)
    ax.text(1.11, -1.00, sx, fontsize=8)
    ax.text(1.11, -1.07, sy, fontsize=8)
    fig.savefig(fname+'.eps', format='eps', dpi=1200)
    plt.show()
    
    
def plotSkyTadpoles(hhead, vhead, htail, vtail, magnif, parms, rmsx, rmsy, fname, note):
    dh = htail - hhead   # horiz axis
    dv = vtail - vhead   # vertical axis
    hmagtail = hhead + magnif*dh
    vmagtail = vhead + magnif*dv
    fig = plt.figure(figsize = (6.5, 6.5))          # 6.5 inches =max LaTeX.
    ax = fig.add_axes([0.3, 0.1, 0.6, 0.6])         # left, bottom, width, height,
    ax.plot(hhead, vhead, 'ko', markersize=2) # dots
    ax.plot([hhead,hmagtail],[vhead,vmagtail],'r')  # line segments
    ax.set_xlim(-1.1, +1.1)
    ax.set_ylim(-1.1, +1.1)
    # plt.axis('equal')  # spoils the specified limits!
    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.text(-1.05, -0.98, 'TadpoleMag = '+str(int(magnif)), fontsize=8)
    ax.text(-1.05, -1.03, 'Heads: {x,y} from sky grid', fontsize=8)
    ax.text(-1.05, -1.08, 'Tails: fit to '+fname, fontsize=8)
    ax.text(-1.05, +1.00, fiducials, fontsize=8)
    ax.text(-1.05, +0.95, note, fontsize=8)
    ax.text(+0.55, -1.08, PROGNAME, fontsize=8)
    plt.title("{X,Y} fit deviations from actual {X,Y}") 
    for i in range(0, NPARMS):
        sparm = ' {:2d}  {:9.6f}'.format(i, parms[i])
        y = 1.15 - 0.07*i
        ax.text(-1.95, y, sparm, fontsize=6)
    sx = ' rmsX={:9.6f}'.format(rmsx)
    sy = ' rmsY={:9.6f}'.format(rmsy)
    total = np.sqrt(rmsx**2 + rmsy**2)
    st = ' both={:9.6f}'.format(total)
    ax.text(-1.95, -1.00, sx, fontsize=8)
    ax.text(-1.95, -1.07, sy, fontsize=8)
    ax.text(-1.95, -1.14, st, fontsize=8)    
    fig.savefig(fname+'.eps', format='eps', dpi=600)
    plt.show()
    






def anyFunc(uv, *parms): # uv = 1D array input concatenate(u[],v[])
    # Model: a sum of polynomials
    # note: curve_fit() calls its given func(args, *parms), always args then parms
    # print "anyFunc() has received parms = "
    # print ["{0:0.6f}".format(xx) for xx in parms]
    pq = np.zeros(len(uv))
    half = len(uv)/2
    for i in range(0, half):  # EACH OBJECT POINT NOT AN ARRAY
        u = uv[i]
        v = uv[i+half]
        x, y = getSum(u, v, *parms)  # always args then parms
        pq[i] = x
        pq[i+half] = y
    return pq   

def getSum(u, v, *coefs):
    # This models the distortion of ONE SKY POINT onto ONE LABFRAME FOCAL POINT.
    # Sums over all terms.
    # coefs is the array of NPARMS coefficients
    # u and v are Cartesian sky coordinates mapped into unit radius circle.
    x = u   # initially undeviated
    y = v   # initially undeviated
    for index in range(0, len(coefs)):
        dx, dy = getLab(u, v, index)  # always args then parms
        x += dx * coefs[index]
        y += dy * coefs[index]
    return x, y


     
#-------------------Main program-------------------------    

"""
#-------------analytic orthogonality test-----------------
print '\n\n Analytic Ortho dot Ortho:'
for row in range(0, NROWS):
    for col in range(0, NROWS):
        dot = getOrthoDot(row, col)
        print '{:8.5f}'.format(dot),
    print ''    
     
#-----------------numerical orthogonality test------------------
print '\n\nStarting numerical orthogonality test'
def buildCircles(radius, ncircles):
    edgegap = 0.47
    radius = radius*(1. - edgegap/ncircles)
    pairs = list()
    pairs.append([0.,0.])
    count = 1
    for icirc in range(1, ncircles+1):
        daz = 60./icirc
        off = 0.0 if icirc%2==0 else daz/2.0
        r = icirc * radius/ncircles
        for jaz in range(0, 6*icirc):
            a = np.deg2rad(off + jaz*daz)
            x = r * np.cos(a)
            y = r * np.sin(a)
            pairs.append([x,y])
            count += 1
    print "#  buildCircles() ncircles has created npairs = ", ncircles, count
    return np.array(pairs)

args = buildCircles(1.00, 48)  # 18 gets 1027 points;  48 gets 7057
NPTS = len(args)

norms = np.zeros(NPARMS)  # preliminary normalization
for which in range(0, NPARMS):
    sum = 0.
    for i in range(0, NPTS):
        x, y = getLocal(args[i,0], args[i,1], which)
        sum += x*x + y*y
    norms[which] = np.sqrt(NPTS/sum)
    print 'Norm for which = ', which, norms[which], norms[which]**2

funcs = np.zeros((NPARMS, NPTS, 2))
for which in range(0, NPARMS):
    for i in range(0, NPTS):
        x, y = getLocal(args[i,0], args[i,1], which)
        funcs[which, i, 0] = norms[which] * x
        funcs[which, i, 1] = norms[which] * y
print "\nCreated the list of funcs."     

dotproducts = np.zeros((NPARMS, NPARMS))
for which1 in range(0, NPARMS):
    for which2 in range(0, NPARMS):
        xx = 0.
        yy = 0.
        for i in range(0, NPTS):
            xx += funcs[which1, i, 0] * funcs[which2, i, 0]   
            yy += funcs[which1, i, 1] * funcs[which2, i, 1]
        dotproducts[which1, which2] = xx + yy
print "Created matrix of numerically normalized dot products:"
for which1 in range(0, NPARMS):
    print '#  ',
    for which2 in range(0, NPARMS):
        print '{:6.3f}'.format(dotproducts[which1, which2]/NPTS),
    print ''
"""




#---load the fiducial coordinates------------
print 'Loading fiducials file name = ', fiducials




#--------------begin the run sequence-----------
#---strange: run zero has expected stats but subsequent runs do not.

pinitial = np.zeros(NPARMS)
bins = np.zeros(NBINS)   # gather deviations from all trials; values range from -100 to +100 ppm  


for trial in range(0, NTRIALS):
    print '\nStarting trial number ', trial 
    Urt, Vrt, Xrt, Yrt = getRaysCSV(fiducials)    # get fiducial ray trace data 
    args = np.concatenate([Urt,Vrt])
    print 'fiducial ray trace data built; len(args) = ', len(args)
    UVrt = args
    if trial==0:
        plotXYtadpoles(Urt,Vrt,Xrt,Yrt, 10.,fiducials)   # V and Y are now vertical upward.
    goals = np.concatenate([Xrt,Yrt])
    goals += 1E-6 * rmsppm * np.random.normal(0., 1., len(goals))
    print 'Starting the fiducials fit with len(goals) = ', len(goals)
    popt, covar = curve_fit(anyFunc, args, goals, pinitial)
    for which in range(0, NPARMS):  # list the best fit parameters
        x, y = getLab(0., 0., which) 
        print which, '  {:12.6f}'.format(popt[which])
    XYfit = anyFunc(args, *popt)
    xylen = len(XYfit)
    print 'fitted function XYfit has length = ', len(XYfit)
    half = xylen/2
    Xfit = XYfit[0:half]
    Yfit = XYfit[half:xylen]
    sosx = 0.
    sosy = 0.
    for i in range(0, len(Xrt)):
       sosx += (Xfit[i] - Xrt[i])**2
       sosy += (Yfit[i] - Yrt[i])**2
    rmsx = np.sqrt(sosx / len(Xrt))
    rmsy = np.sqrt(sosy / len(Xrt))
    print "fit rmsx = ", rmsx
    print "fit rmsy = ", rmsy
    # print "NPARMS = ", NPARMS
    if trial==0:
        plotFitTadpoles(Xrt, Yrt, Xfit, Yfit, tadpole, popt, rmsx, rmsy, fiducials, plotnote)
        
    # Now use these parms to interpolate the big sky grid.
    Urt, Vrt, Xrt, Yrt = getRaysCSV(bigskygrid)  # returns normalized sky ray trace data
    args = np.concatenate([Urt,Vrt])             # convert to 1D
    print 'sky ray trace data built; len(args) = ', len(args)
    XYtest = anyFunc(args, *popt)                # generate the distorted map
    xylen = len(XYtest)
    half = xylen/2
    Xtest = XYtest[0:half]
    Ytest = XYtest[half:xylen]
    sosx = 0.
    sosy = 0.
    for i in range(0, len(Xrt)):
        dx = Xtest[i] - Xrt[i]
        dy = Ytest[i] - Yrt[i]
        xbin = int(NBINS/2 + 1E6*dx)
        if xbin>0 and xbin<NBINS:
            bins[xbin] += 1
        ybin = int(NBINS/2 + 1E6*dy)
        if ybin>0 and ybin<NBINS:
            bins[ybin] += 1
        # print '{:6d}{:6d}'.format(xbin, ybin)   # diagnostic
        sosx += dx**2
        sosy += dy**2    
    rmsx = np.sqrt(sosx / len(Xrt))
    rmsy = np.sqrt(sosy / len(Xrt))
    print "sky rmsx = ", rmsx
    print "sky rmsy = ", rmsy
    # print "NPARMS = ", NPARMS
    if trial==0:
        plotSkyTadpoles(Xrt, Yrt, Xtest, Ytest, tadpole, popt, rmsx, rmsy, bigskygrid, plotnote)
    print 'plotnote = ', plotnote

"""
#---finally run a bunch of randomized variants of the fit + sky, and write the errors----
for trial in range(0, NTRIALS):    
    args = np.concatenate([Urt,Vrt])
    UVrt = args
    goals = np.concatenate([Xrt,Yrt])
    goals += 1E-6 * rmsppm * np.random.normal(0., 1., len(goals))
    #---1E-6 because rmsppm is in parts per million "ppm"--------

    #--DO THE FIT-----------

    popt, covar = curve_fit(anyFunc, args, goals, pinitial)

    #---Try out the popt[] interpolation parms on a sky grid-------
    #------ load Urt, Vrt, Xrt, Yrt from ray tracing------------------

    getRaysCSV(bigskygrid)     # produces normalized Urt, Vrt, Xrt, Yrt
    args = np.concatenate([Urt,Vrt])             # convert to 1D
    XYtest = anyFunc(args, *popt)                # generate the distorted map
    xylen = len(XYtest)
    half = xylen/2
    Xtest = XYtest[0:half]
    Ytest = XYtest[half:xylen]
    sosx = 0.
    sosy = 0.
    for i in range(0, len(Xrt)):
        dx = Xtest[i] - Xrt[i]
        dy = Ytest[i] - Yrt[i]
        xbin = int(NBINS/2 + 1E6*dx)
        if xbin>0 and xbin<NBINS:
            bins[xbin] += 1
        ybin = int(NBINS/2 + 1E6*dy)
        if ybin>0 and ybin<NBINS:
            bins[ybin] += 1
        print '{:6d}{:6d}'.format(xbin, ybin)   # diagnostic
        sosx += dx**2
        sosy += dy**2
    rmsx = np.sqrt(sosx / len(Xrt))
    rmsy = np.sqrt(sosy / len(Xrt))
    print "rmsx = ", rmsx
    print "rmsy = ", rmsy
    result = '{:4d},{:12.6f},{:12.6f}\n'.format(trial, dx, dy)
    print result
    outfile.write(result)
    outfile.write('\n')
"""    
    
outfile = open('120fid-L16ortho2-100ppm.txt', 'w')
for bindex in range(0, NBINS):
    howmany = bins[bindex]
    result = '{:6d} {:10.0f}\n'.format(bindex, howmany)
    outfile.write(result) 
outfile.close()

    
    
